#include <std_include.hpp>
#include "loader/component_loader.hpp"

#include <utils/cryptography.hpp>
#include <utils/hook.hpp>
#include <utils/io.hpp>
#include <utils/nt.hpp>
#include <utils/properties.hpp>
#include <utils/smbios.hpp>
#include <utils/string.hpp>

#include "auth.hpp"

namespace auth {
namespace {
std::string get_hw_profile_guid() {
  HW_PROFILE_INFO info;
  if (!GetCurrentHwProfileA(&info)) {
    return {};
  }

  return std::string{info.szHwProfileGuid, sizeof(info.szHwProfileGuid)};
}

std::string get_protected_data() {
  std::string input = "Alter-Ware-IW4x-SP-Auth";

  DATA_BLOB data_in{}, data_out{};
  data_in.pbData = reinterpret_cast<std::uint8_t*>(input.data());
  data_in.cbData = static_cast<DWORD>(input.size());
  if (CryptProtectData(&data_in, nullptr, nullptr, nullptr, nullptr,
                       CRYPTPROTECT_LOCAL_MACHINE, &data_out) != TRUE) {
    return {};
  }

  const auto size = std::min<std::uint32_t>(data_out.cbData, 52);
  std::string result(reinterpret_cast<char*>(data_out.pbData), size);
  LocalFree(data_out.pbData);

  return result;
}

std::string get_hdd_serial() {
  DWORD serial{};
  if (!GetVolumeInformationA("C:\\", nullptr, 0, &serial, nullptr, nullptr,
                             nullptr, 0)) {
    return {};
  }

  return utils::string::va("{0:08X}", serial);
}

std::string get_key_entropy() {
  std::string entropy{};
  entropy.append(utils::smbios::get_uuid());
  entropy.append(get_hw_profile_guid());
  entropy.append(get_protected_data());
  entropy.append(get_hdd_serial());

  if (entropy.empty()) {
    entropy.resize(32);
    utils::cryptography::random::get_data(entropy.data(), entropy.size());
  }

  return entropy;
}

bool load_key(utils::cryptography::ecc::key& key) {
  std::string data{};

  auto key_path = (utils::properties::get_appdata_path() / "iw4xsp-private.key")
                      .generic_string();
  if (!utils::io::read_file(key_path, &data)) {
    return false;
  }

  key.deserialize(data);
  if (!key.is_valid()) {
    game::Com_PrintError(game::CON_CHANNEL_ERROR, "Loaded key is invalid!\n");
    return false;
  }

  return true;
}

utils::cryptography::ecc::key generate_key() {
  auto key = utils::cryptography::ecc::generate_key(512, get_key_entropy());
  if (!key.is_valid()) {
    throw std::runtime_error("Failed to generate cryptographic key!");
  }

  auto key_path = (utils::properties::get_appdata_path() / "iw4xsp-private.key")
                      .generic_string();
  if (!utils::io::write_file(key_path, key.serialize())) {
    game::Com_PrintError(game::CON_CHANNEL_ERROR,
                         "Failed to write cryptographic key!\n");
  }

  game::Com_PrintError(game::CON_CHANNEL_ERROR,
                       "Generated cryptographic key: %llX\n", key.get_hash());
  return key;
}

utils::cryptography::ecc::key load_or_generate_key() {
  utils::cryptography::ecc::key key{};
  if (load_key(key)) {
    game::Com_PrintError(game::CON_CHANNEL_ERROR,
                         "Loaded cryptographic key: %llX\n", key.get_hash());
    return key;
  }

  return generate_key();
}

utils::cryptography::ecc::key get_key_internal() {
  auto key = load_or_generate_key();

  auto key_path = (utils::properties::get_appdata_path() / "iw4xsp-public.key")
                      .generic_string();
  if (!utils::io::write_file(key_path, key.get_public_key())) {
    game::Com_PrintError(game::CON_CHANNEL_ERROR,
                         "Failed to write public key!\n");
  }

  return key;
}

const utils::cryptography::ecc::key& get_key() {
  static auto key = get_key_internal();
  return key;
}

bool is_second_instance() {
  static const auto is_first = []() -> bool {
    static utils::nt::handle mutex =
        CreateMutexA(nullptr, FALSE, "iw4xsp_mutex");
    return mutex && GetLastError() != ERROR_ALREADY_EXISTS;
  }();

  return !is_first;
}
} // namespace

std::uint64_t get_guid() {
  static const auto guid = []() -> std::uint64_t {
    if (is_second_instance()) {
      return 0x110000100000000 |
             (::utils::cryptography::random::get_integer() & ~0x80000000);
    }

    return get_key().get_hash();
  }();

  return guid;
}

class component final : public component_interface {
public:
  void post_load() override {
    // Patch Steam_GetClientIDAsXUID
    utils::hook::set<std::uint32_t>(0x4911B0, 0xC301B0);
  }
};
} // namespace auth

REGISTER_COMPONENT(auth::component)
